#include "time/RosenbrockAdaptInstationary.h"
#include "AdaptInfo.h"
#include "ProblemTimeInterface.h"

namespace AMDiS {

  RosenbrockAdaptInstationary::RosenbrockAdaptInstationary(std::string name, 
							   RosenbrockStationary *problemStat,
							   AdaptInfo *info,
							   ProblemTimeInterface *problemInstat,
							   AdaptInfo *initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      tauGamma(1.0)
  {
    std::string str;
    GET_PARAMETER(0, name + "->rosenbrock method", &str);
    RosenbrockMethodCreator *creator = 
      dynamic_cast<RosenbrockMethodCreator*>(CreatorMap<RosenbrockMethod>::getCreator(str));
    rosenbrockMethod = creator->create();

    problemStat->setRosenbrockMethod(rosenbrockMethod);
  }


  void RosenbrockAdaptInstationary::oneTimestep()
  {
    FUNCNAME("RosenbrockAdaptInstationary::oneTimestep()");

    // estimate before first adaption
    if (adaptInfo->getTime() <= adaptInfo->getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);

    bool rejected = false;
    double timeTol = adaptInfo->getTimeTolerance(0);
    
    do {
      // increment time
      adaptInfo->setTime(adaptInfo->getTime() + adaptInfo->getTimestep());
      problemTime->setTime(adaptInfo);
      tauGamma = 1.0 / (adaptInfo->getTimestep() * rosenbrockMethod->getGamma());
      
      INFO(info, 6)("time = %e   timestep = %e\n",
		    adaptInfo->getTime(), adaptInfo->getTimestep());

      adaptInfo->setSpaceIteration(0);
      
      // do the iteration
      problemIteration->beginIteration(adaptInfo);
      problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
      problemIteration->endIteration(adaptInfo);

      double errorEst = adaptInfo->getTimeEstSum(0);
      double newTimestep = 0.0;
      double order = rosenbrockMethod->getOrder();

      if (errorEst < timeTol) {
	double fac = 1.0;

	if (firstTimestep || succRejection) {
	  fac = pow((timeTol / errorEst), 1.0 / order);
	} else {
	  fac = adaptInfo->getTimestep() / tauAcc * 
	    pow((timeTol * estAcc / (errorEst * errorEst)), 1.0 / order);
	}
	fac = std::min(fac, 3.0);
	newTimestep = fac * adaptInfo->getTimestep();

	tauAcc = adaptInfo->getTimestep();
	estAcc = errorEst;

	lastTimestepRejected = false;
	succRejection = false;
      } else {
	if (lastTimestepRejected) {
	  succRejection = true;  

	  double reducedP = 
	    log(errorEst / estRej) / log(adaptInfo->getTimestep() / tauRej);
	  
	  if (reducedP < order && reducedP > 0.0)
	    order = reducedP;
	} 

	newTimestep = pow((timeTol / errorEst), 1.0 / order) * adaptInfo->getTimestep();
	
	tauRej = adaptInfo->getTimestep();
	estRej = errorEst;

	lastTimestepRejected = true;
      }
	
      newTimestep *= 0.95;

      if (errorEst < timeTol || firstTimestep) {
	INFO(info, 6)("Accepted timestep at time = %f   with timestep = %f\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());

	adaptInfo->setTimestep(newTimestep);

	if (firstTimestep)
	  firstTimestep = false;

	rejected = false;

	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimator for component %d = %f\n", 
			i, adaptInfo->getTimeEstSum(i));
      } else {
	INFO(info, 6)("Rejected timestep at time = %f   with timestep = %f\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());
	
	adaptInfo->setTime(adaptInfo->getTime() - adaptInfo->getTimestep());
	adaptInfo->setTimestep(newTimestep);

	rejected = true;
      }

    } while (rejected);

    rosenbrockStat->acceptTimestep();

    adaptInfo->setLastProcessedTimestep(adaptInfo->getTimestep());
    adaptInfo->incTimestepNumber();
  }

}
